/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* Copyright 2013 - 2020, nymea GmbH
* Contact: contact@nymea.io
*
* This file is part of maveo-scanner.
* This project including source code and documentation is protected by
* copyright law, and remains the property of nymea GmbH. All rights, including
* reproduction, publication, editing and translation, are reserved. The use of
* this project is subject to the terms of a license agreement to be concluded
* with nymea GmbH in accordance with the terms of use of nymea GmbH, available
* under https://nymea.io/license
*
* For any further details and any questions please contact us under
* contact@nymea.io or see our FAQ/Licensing Information on
* https://nymea.io/license/faq
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef BOOTLOADERHANDLER_H
#define BOOTLOADERHANDLER_H

#include <QObject>
#include <QSerialPort>

#include "firmwarefile.h"
#include "bootloaderutils.h"

// https://www.cypress.com/documentation/application-notes/an68272-psoc-3-psoc-4-psoc-5lp-and-psoc-analog-coprocessor-uart

class LIBMAVEO_SCANNER_EXPORT BootloaderHandler : public QObject
{
    Q_OBJECT
public:
    enum State {
        StateIdle,
        StateConnecting,
        StateEnterBootloader,
        StateGetFlahSize,
        StatePrepareUpload,
        StateStreamRow,
        StateProgramRow,
        StateVerifyRow,
        StateVerifyApplication,
        StateExitBootloader
    };
    Q_ENUM(State)

    explicit BootloaderHandler(QObject *parent = nullptr);

    void setPortName(const QString &portName);
    void setBaudrate(qint32 baudrate);

    void setFirmwareData(const FirmwareData &firmwareData);

    State state() const;
    double percentage() const;

private:
    State m_state = StateIdle;

    QSerialPort *m_serialPort = nullptr;
    QString m_portName;
    qint32 m_baudrate = QSerialPort::Baud115200;
    bool m_connected = false;

    QByteArray m_buffer;
    int m_bufferCurrentMessageSize = -1;

    // Bootloader information
    quint32 m_siliconId = 0;
    quint8 m_siliconRev = 0;
    quint32 m_bootloaderVersion = 0;
    quint16 m_startRow = 0;
    quint16 m_endRow = 0;

    // Update file information
    FirmwareData m_firmwareData;
    int m_currentUploadTotalSize = 0;
    int m_currentUploadProgress = 0;
    double m_percentage = 0;

    int m_currentUploadRowIndex = 0;
    int m_currentUploadRowOffset = 0;
    int m_currentUploadRowPayloadSize = 0;

    QByteArray buildPackage(BootloaderUtils::BootloaderCommand command, const QByteArray &payload = QByteArray());

    void setState(State state);
    void setConnected(bool connected);

    void processData(const QByteArray &data);
    void processMessage(BootloaderUtils::BootloaderError errorCode, const QByteArray &payload);

    FirmwareDataRow getCurrentRow() const;
    void calculateUpdatePercentage();

    void sendData(const QByteArray &data);
    void sendNextRowData();

    void reset();

signals:
    void stateChanged(State state);
    void connectedChanged(bool connected);
    void percentageChanged(double percentage);
    void finished(bool success);

private slots:
    void onReadyRead();
    void onError(QSerialPort::SerialPortError error);
    void onExitBootloaderTimeout();

public slots:
    void start();
    void stop();
};

#endif // BOOTLOADERHANDLER_H
