/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* Copyright 2013 - 2020, nymea GmbH
* Contact: contact@nymea.io
*
* This file is part of maveo-scanner.
* This project including source code and documentation is protected by
* copyright law, and remains the property of nymea GmbH. All rights, including
* reproduction, publication, editing and translation, are reserved. The use of
* this project is subject to the terms of a license agreement to be concluded
* with nymea GmbH in accordance with the terms of use of nymea GmbH, available
* under https://nymea.io/license
*
* For any further details and any questions please contact us under
* contact@nymea.io or see our FAQ/Licensing Information on
* https://nymea.io/license/faq
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef HONEYWELLSCANNER_H
#define HONEYWELLSCANNER_H

#include <QQueue>
#include <QObject>

#include "maveoscannerutils.h"

class HoneywellReply;
class HoneywellRequest;

class LIBMAVEO_SCANNER_EXPORT HoneywellScanner : public QObject
{
    Q_OBJECT
    Q_PROPERTY(TriggerMode triggerMode READ triggerMode NOTIFY triggerModeChanged)
    Q_PROPERTY(QString vendorName READ vendorName NOTIFY vendorNameChanged)
    Q_PROPERTY(QString productName READ productName NOTIFY productNameChanged)
    Q_PROPERTY(QString serialNumber READ serialNumber NOTIFY serialNumberChanged)
    Q_PROPERTY(QString firmwareRevision READ firmwareRevision NOTIFY firmwareRevisionChanged)
    Q_PROPERTY(QString projectRevision READ projectRevision NOTIFY projectRevisionChanged)
    Q_PROPERTY(QString decoderRevision READ decoderRevision NOTIFY decoderRevisionChanged)

    friend class MaveoScanner;

public:
    enum State {
        StateUninitialized,
        StateInitializing,
        StateReady,
        StateError
    };
    Q_ENUM(State)

    enum Response {
        ResponseUnknown = -1,
        ResponseEnq = 0x05,
        ResponseAck = 0x06,
        ResponseNak = 0x15
    };
    Q_ENUM(Response)

    enum TriggerMode {
        TriggerModeManual,
        TriggerModePresentation
    };
    Q_ENUM(TriggerMode)

    enum Error {
        ErrorNoError,
        ErrorReplyTimeout,
        ErrorReadingDeviceInformation,
        ErrorSettingConfiguration
    };
    Q_ENUM(Error)

    explicit HoneywellScanner(QObject *parent = nullptr);

    State state() const;

    Error error() const;

    // When this mode is selected, your scan engine is optimized to read bar codes from mobile phone or other LED displays. How-
    // ever, the speed of scanning printed bar codes may be slightly lower when this mode is enabled.
    bool mobilePhoneReadModeEnabled() const;

    TriggerMode triggerMode() const;

    QString vendorName() const;
    QString productName() const;
    QString serialNumber() const;

    QString firmwareRevision() const;
    QString projectRevision() const;
    QString decoderRevision() const;

    Q_INVOKABLE HoneywellReply *setTriggerMode(TriggerMode triggerMode);
    Q_INVOKABLE HoneywellReply *setMobilePhoneReadModeEnabled(bool mobilePhoneReadModeEnabled);

    // Commands without response
    Q_INVOKABLE void reset();
    Q_INVOKABLE void trigger();
    Q_INVOKABLE void untrigger();

private:
    HoneywellReply *m_currentReply = nullptr;
    QQueue<HoneywellReply *> m_replies;

    State m_state = StateUninitialized;
    TriggerMode m_triggerMode = TriggerModePresentation;
    bool m_mobilePhoneReadModeEnabled = true;
    Error m_error = ErrorNoError;

    QString m_vendorName;
    QString m_productName;
    QString m_serialNumber;

    QString m_firmwareRevision;
    QString m_projectRevision;
    QString m_decoderRevision;

    void initialize();

    void setState(State state);
    void setError(Error error);

    void resetData();

    HoneywellReply *sendRequest(const QString &command, char target = 'm', char join = '\r');

signals:
    void requestDataToSend(const QByteArray &data);
    void stateChanged(HoneywellScanner::State state);
    void errorOccurred(HoneywellScanner::Error error);
    void ready();

    void vendorNameChanged(const QString &vendorName);
    void productNameChanged(const QString &productName);
    void serialNumberChanged(const QString &serialNumber);
    void firmwareRevisionChanged(const QString &firmwareRevision);
    void projectRevisionChanged(const QString &projectRevision);
    void decoderRevisionChanged(const QString &decoderRevision);

    void codeScanned(const QString &code);
    void triggerModeChanged(HoneywellScanner::TriggerMode triggerMode);
    void mobilePhoneReadModeEnabledChanged(bool mobilePhoneReadModeEnabled);

private slots:
    void sendNextReply();
    void initializeInternally();

public slots:
    void onConnectedChanged(bool connected);
    void processData(const QByteArray &data);

};

#endif // HONEYWELLSCANNER_H
