/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* Copyright 2013 - 2020, nymea GmbH
* Contact: contact@nymea.io
*
* This file is part of maveo-scanner.
* This project including source code and documentation is protected by
* copyright law, and remains the property of nymea GmbH. All rights, including
* reproduction, publication, editing and translation, are reserved. The use of
* this project is subject to the terms of a license agreement to be concluded
* with nymea GmbH in accordance with the terms of use of nymea GmbH, available
* under https://nymea.io/license
*
* For any further details and any questions please contact us under
* contact@nymea.io or see our FAQ/Licensing Information on
* https://nymea.io/license/faq
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef MAVEOSCANNER_H
#define MAVEOSCANNER_H

#include <QObject>
#include <QColor>
#include <QTimer>

#include "honeywellreply.h"
#include "honeywellscanner.h"
#include "maveoscannerreply.h"
#include "maveoscannerinterface.h"
#include "firmware/bootloaderhandler.h"
#include "firmware/firmwaredownloader.h"

class MaveoScannerPackage;

class LIBMAVEO_SCANNER_EXPORT MaveoScanner : public QObject
{
    Q_OBJECT
    Q_PROPERTY(FirmwareDownloader *firmwareDownloader READ firmwareDownloader CONSTANT)
    Q_PROPERTY(HoneywellScanner *honeywellScanner READ honeywellScanner CONSTANT)

    Q_PROPERTY(State state READ state NOTIFY stateChanged)
    Q_PROPERTY(QString portName READ portName CONSTANT)
    Q_PROPERTY(qint32 baudrate READ baudrate CONSTANT)

    Q_PROPERTY(QString serialPortName READ serialPortName NOTIFY serialPortNameChanged)
    Q_PROPERTY(QString serialPortSystemLocation READ serialPortSystemLocation NOTIFY serialPortSystemLocationChanged)
    Q_PROPERTY(QString serialPortDescription READ serialPortDescription NOTIFY serialPortDescriptionChanged)
    Q_PROPERTY(QString serialPortManufacturer READ serialPortManufacturer NOTIFY serialPortManufacturerChanged)
    Q_PROPERTY(QString serialPortSerialNumber READ serialPortSerialNumber NOTIFY serialPortSerialNumberChanged)
    Q_PROPERTY(quint16 serialPortVendorIdentifier READ serialPortVendorIdentifier NOTIFY serialPortVendorIdentifierChanged)
    Q_PROPERTY(quint16 serialPortProductIdentifier READ serialPortProductIdentifier NOTIFY serialPortProductIdentifierChanged)

    Q_PROPERTY(bool available READ available NOTIFY availableChanged)
    Q_PROPERTY(QString firmwareVersion READ firmwareVersion NOTIFY firmwareVersionChanged)

    Q_PROPERTY(double updatePercentage READ updatePercentage NOTIFY updatePercentageChanged)
    Q_PROPERTY(bool firmwareUpdateAvailable READ firmwareUpdateAvailable NOTIFY firmwareUpdateAvailableChanged)

public:
    enum State {
        StateUnconnected,
        StateConnecting,
        StateConnected,
        StateConnectingBootloader,
        StateConnectedBootloader
    };
    Q_ENUM(State)

    // Default constructor if the serial port is known
    explicit MaveoScanner(QObject *parent = nullptr);
    explicit MaveoScanner(const QString &serialNumber, QObject *parent = nullptr);
    explicit MaveoScanner(MaveoScannerDeviceWatcher *deviceWatcher, QObject *parent = nullptr);
    explicit MaveoScanner(MaveoScannerDeviceWatcher *deviceWatcher, const QString &serialNumber, QObject *parent = nullptr);

    State state() const;

    MaveoScannerInterface *scannerInterface() const;
    FirmwareDownloader *firmwareDownloader() const;
    HoneywellScanner *honeywellScanner() const;

    // Port information
    QString portName() const;
    void setPortName(const QString &portName);

    qint32 baudrate() const;
    void setBaudrate(qint32 baudrate);

    QString serialPortName() const;
    QString serialPortSystemLocation() const;
    QString serialPortDescription() const;
    QString serialPortManufacturer() const;
    QString serialPortSerialNumber() const;
    quint16 serialPortVendorIdentifier() const;
    quint16 serialPortProductIdentifier() const;

    // Scanner information
    bool available() const;
    QString firmwareVersion() const;

    // Firmware update
    bool firmwareUpdateAvailable() const;
    double updatePercentage() const;

    // Scanner methods
    Q_INVOKABLE void resetController();
    Q_INVOKABLE void startFirmwareUpdate(const QString &firmwareFile = QString());
    Q_INVOKABLE void setColor(quint8 command, uint duration, const QColor &color);
    Q_INVOKABLE void playSound(quint8 command, uint octave, uint note, uint volume, uint duration);
    Q_INVOKABLE void setRelay(bool power);
    Q_INVOKABLE void goodReadFeedback();
    Q_INVOKABLE void badReadFeedback();
    Q_INVOKABLE void playTestSound();

private:
    QTimer *m_heartbeatTimer = nullptr;

    State m_state = StateUnconnected;
    QString m_portName;
    qint32 m_baudrate = 115200;
    bool m_available = false;
    bool m_scannerReady = false;
    bool m_initializing = false;
    bool m_firmwareUpdateAvailable = false;

    MaveoScannerInterface *m_interface = nullptr;
    HoneywellScanner *m_honeywellScanner = nullptr;
    BootloaderHandler *m_bootloaderHandler = nullptr;
    FirmwareDownloader *m_firmwareDownloader = nullptr;

    void construct(MaveoScannerDeviceWatcher *deviceWatcher);

    // Port infos
    QString m_serialPortName;
    QString m_serialPortSystemLocation;
    QString m_serialPortDescription;
    QString m_serialPortManufacturer;
    QString m_serialPortSerialNumber;
    quint16 m_serialPortVendorIdentifier = 0;
    quint16 m_serialPortProductIdentifier = 0;

    // Protocol data
    quint8 m_messageId = 0;
    QString m_firmwareVersion;

    QHash<quint8, MaveoScannerReply *> m_runningRequests;

    void setState(State state);
    void setAvailable(bool available);
    void setFirmwareVersion(const QString &firmwareVersion);
    void setFirmwareUpdateAvailable(bool firmwareUpdateAvailable);

    void setSerialPortName(const QString &serialPortName);
    void setSerialPortSystemLocation(const QString &serialPortSystemLocation);
    void setSerialPortDescription(const QString &serialPortDescription);
    void setSerialPortManufacturer(const QString &serialPortManufacturer);
    void setSerialPortSerialNumber(const QString &serialPortSerialNumber);
    void setSerialPortVendorIdentifier(quint16 serialPortVendorIdentifier);
    void setSerialPortProductIdentifier(quint16 serialPortProductIdentifier);

    int m_initRetry = 0;
    void initScanner();
    void runInitProcedure();
    void verifyInitProcedureRetry();

    MaveoScannerReply *createReply(quint8 messageId);

    // Internal firmware calls
    MaveoScannerReply *sendScannerData(const QByteArray &data);
    MaveoScannerReply *sendHeartbeat();
    MaveoScannerReply *sendReadFirmwareVersion();
    MaveoScannerReply *sendResetController();
    MaveoScannerReply *sendGetStatus();
    MaveoScannerReply *sendPlayTestSound();

    MaveoScannerReply *sendStartBootloader();
    MaveoScannerReply *sendSetLedCommand(MaveoScannerUtils::LedCommand command, uint duration, const QColor &color);
    MaveoScannerReply *sendPlaySoundCommand(MaveoScannerUtils::BuzzerCommand command, MaveoScannerUtils::Octave octave, MaveoScannerUtils::Note note, MaveoScannerUtils::Volume volume, uint duration);
    MaveoScannerReply *sendSetRelayCommand(bool power);
    MaveoScannerReply *sendGoodReadCommand();
    MaveoScannerReply *sendBadReadCommand();

private slots:
    void onHeartbeatTimeout();
    void onInterfaceConnectedChanged(bool connected);
    void onInterfaceDataReceived(const QByteArray &data);
    void onInterfacePackageReceived(const MaveoScannerPackage &package);

    void onHoneywellStateChanged(HoneywellScanner::State state);
    void onHoneywellDataSendRequest(const QByteArray &data);

    void onBootloaderHandlerStateChaned(BootloaderHandler::State state);
    void onBootloaderHandlerConnectedChanged(bool connected);
    void onBootloaderHandlerFinished(bool success);

    void onFirmwareDownloaderFirmwareChanged(const FirmwareData &firmware);

signals:
    void stateChanged(State state);
    void availableChanged(bool available);
    void firmwareVersionChanged(const QString &firmwareVersion);
    void firmwareUpdateAvailableChanged(bool firmwareUpdateAvailable);
    void updatePercentageChanged(double updatePercentage);

    void serialPortNameChanged(const QString &serialPortName);
    void serialPortSystemLocationChanged(const QString &serialPortSystemLocation);
    void serialPortDescriptionChanged(const QString &serialPortDescription);
    void serialPortManufacturerChanged(const QString &serialPortManufacturer);
    void serialPortSerialNumberChanged(const QString &serialPortSerialNumber);
    void serialPortVendorIdentifierChanged(quint16 serialPortVendorIdentifier);
    void serialPortProductIdentifierChanged(quint16 serialPortProductIdentifier);

    void codeScanned(const QString &code);

public slots:
    bool start();
    void stop();

};

#endif // MAVEOSCANNER_H
